module ReadFRRaw (itemStream) where

import Data.List
import Text.Regex
import ReadFR


itemStream :: String -> IO [FRItem]
itemStream fileName =
	do
		contents <- readFile fileName
		return $ seeHeader $ skipEltsHead $ lines contents


skipEltsHead :: ([String] -> [String])
skipEltsHead = dropWhile (== "") . tail . dropWhile (/= "-->")


cateStart, itemStart, attribute, header :: Regex
cateStart = mkRegex "^<category name=\"([^\"]*)\""
itemStart = mkRegex "^<item( .*)>$"
attribute = mkRegex " ([^=]*)=\"([^\"]*)\"(.*)$"
header = mkRegex "<fullrecall core_version=\"([0-9]*)\" time_of_start=\"([0-9]*)\">$"


seeHeader :: [String] -> [FRItem]
seeHeader (firstLine:rest) =
	case matchRegex header firstLine of
	Just ["12",_] ->
		takeItems [] [] rest
	Just [version,_] ->
		error $ "Unknown version " ++ version
	Just _ ->
		badLine firstLine
	Nothing ->
		badLine firstLine
seeHeader [] = error "header expected but no input"


takeItems :: [FRItem] -> [String] -> [String] -> [FRItem]

takeItems listSoFar _ [] = listSoFar

takeItems listSoFar cateStack ("<fullrecall core_version=\"12\" time_of_start=\"1100741400\">":rest) =
	takeItems listSoFar cateStack rest

takeItems listSoFar cateStack ("":rest) =
	takeItems listSoFar cateStack rest

takeItems listSoFar [] ["</fullrecall>"] = listSoFar

takeItems listSoFar (_:xs) ("</category>":rest) =
	takeItems listSoFar xs rest

takeItems listSoFar cateStack ("</item>":rest) =
	takeItems listSoFar cateStack rest

takeItems listSoFar cateStack (firstLine:rest) =
	case matchRegex cateStart firstLine of
	Just [theName] ->
		takeItems listSoFar (theName : cateStack) rest
	Just _ ->
		badLine firstLine
	Nothing ->
		case matchRegex itemStart firstLine of
		Just [attrs] ->
			(expectQuestion (splitAttrs attrs [])) listSoFar cateStack rest
		Just _ ->
			badLine firstLine
		Nothing ->
			badLine firstLine


splitAttrs :: String -> [(String, String)] -> [(String, String)]
splitAttrs "" soFar = soFar
splitAttrs remainingText soFar =
	case matchRegex attribute remainingText of
	Just [tag,value,rest] ->
		splitAttrs rest ((tag,value):soFar)
	Just _ ->
		badLine ('*':remainingText)
	Nothing ->
		badLine ('%':remainingText)

expectQuestion :: [(String, String)] -> [FRItem] -> [String] -> [String] -> [FRItem]
expectQuestion attrList listSoFar cateStack (('<':'q':'>':firstLine):rest) =
	takeQorA questionClose (expectAnswer attrList) [] firstLine listSoFar cateStack rest
expectQuestion _ _ _ (firstLine:_) = error $ "question expected: " ++ firstLine
expectQuestion _ _ _ _ = error "question expected."


expectAnswer :: [(String, String)] -> String -> [FRItem] -> [String] -> [String] -> [FRItem]
expectAnswer attrList qText listSoFar cateStack (('<':'a':'>':firstLine):rest) =
	takeQorA answerClose (expectItemClose attrList qText) [] firstLine listSoFar cateStack rest
expectAnswer _ qText _ _ _ = error $ "expected answer after " ++ qText


expectItemClose :: [(String, String)] -> String -> String -> [FRItem] -> [String] -> [String] -> [FRItem]
expectItemClose attrList qText aText listSoFar cateStack ("</item>":rest) =
	takeItems (FRItem qText aText attrList cateStack : listSoFar) cateStack rest
expectItemClose _ qText _ _ _ _ = error $ "expected close after " ++ qText


questionClose, answerClose :: String
questionClose = "</q>"
answerClose = "</a>"


takeQorA :: String -> (String -> [FRItem] -> [String] -> [String] -> [FRItem]) -> String -> String -> [FRItem] -> [String] -> [String] -> [FRItem]
takeQorA closing continuation partSoFar nextPartLine listSoFar cateStack stuff@(nextLine:rest) =
	if endsWith nextPartLine closing then
		continuation (partSoFar ++ (stripFinal closing nextPartLine)) listSoFar cateStack stuff
	else
		takeQorA closing continuation (partSoFar ++ nextPartLine) ('\n':nextLine) listSoFar cateStack rest
takeQorA _ _ partSoFar _ _ _ _ = error $ "nothing after Q/A: " ++ partSoFar


stripFinal :: String -> String -> String
stripFinal terminal what = take (length what - length terminal) what


endsWith :: String -> String -> Bool
endsWith string ending = (drop (length string - length ending) string == ending)


badLine :: String -> a
badLine line = error $ "Error in line " ++ line
