module ReadFRXML (itemStream) where

import Data.List
import Text.XML.HaXml.Types
import Text.XML.HaXml.SAX
import Text.Regex
import ReadFR


itemStream :: String -> IO [FRItem]
itemStream filePath =
	do
		contents <- readFile filePath
		return $ result (saxParse filePath contents) [] []


type SaxStream = ([SaxElement], Maybe String)


result :: SaxStream -> [FRItem] -> [String] -> [FRItem]

result (_, Just errmsg) _ _ = error $ "error: " ++ errmsg
result ([], Nothing) itemSoFar [] = itemSoFar
result ([], Nothing) _ _ = error "Categories not closed"
result ((first:rest), whatever) itemSoFar cateStack = oneSaxElement (rest, whatever) itemSoFar cateStack first


oneSaxElement :: SaxStream -> [FRItem] -> [String] -> SaxElement -> [FRItem]

oneSaxElement rest itemSoFar cateStack (SaxDocTypeDecl _) = result rest itemSoFar cateStack

oneSaxElement rest itemSoFar cateStack (SaxProcessingInstruction _) = result rest itemSoFar cateStack

oneSaxElement rest itemSoFar cateStack (SaxComment _) = result rest itemSoFar cateStack

oneSaxElement rest itemSoFar cateStack (SaxElementOpen "fullrecall" attrList) =
	if version == "12" then
		result rest itemSoFar cateStack
	else
		error $ "version " ++ version ++ " incorrect"
	where
		version = lookUpStrAttr "core_version" attrList

oneSaxElement rest itemSoFar cateStack (SaxElementOpen "category" attrList) =
	result rest itemSoFar (lookUpStrAttr "name" attrList : cateStack)

oneSaxElement rest itemSoFar cateStack (SaxElementOpen "item" attrList) =
	takeOneElement itemSoFar cateStack (skipClose (afterAnswer (extractItemPart 'q' (skipBlankInXML rest)))) (map reformAttr attrList)
	where
		skipClose (qPart, (aPart, ((SaxElementClose "item":eltSAXlist), whatever))) = (qPart, aPart, (eltSAXlist,whatever))
		skipClose (qPart, (_, _)) = error $ "skipClose at " ++ qPart
		afterAnswer (qString, qRemains) = (qString, (extractItemPart 'a' qRemains))

oneSaxElement _ _ _ (SaxElementOpen theName _) = error $ "unexpected tag <" ++ theName ++ ">"

oneSaxElement rest itemSoFar (_:cateStack) (SaxElementClose "category") = result rest itemSoFar cateStack

oneSaxElement ([], Nothing) itemSoFar [] (SaxElementClose "fullrecall") = itemSoFar

oneSaxElement _ _ _ (SaxElementClose tag) = error $ "Unexpected </" ++ tag ++ ">"

oneSaxElement rest itemSoFar cateStack (SaxCharData rawData) =
	case matchRegex blanks rawData of
	Just _ ->
		result rest itemSoFar cateStack
	Nothing ->
		error $ "Char data\"" ++ rawData ++ "\""
	where
		blanks = mkRegex "^[ \\t\\n\\r]*$"

oneSaxElement _ _ _ (SaxElementTag theName _) = error $ theName ++ "--no content"

oneSaxElement _ _ _ (SaxReference _) = error "Reference encountered, unexpected"


reformAttr :: Attribute -> (String, String)
reformAttr (name, attValue) = (name, extractSingleValue attValue)


skipBlankInXML :: SaxStream -> SaxStream
skipBlankInXML ((SaxCharData rawData:rest), whatever) =
	if isAllBlank rawData then
		(rest, whatever)
	else
		error $ "Char data \"" ++ rawData ++ "\""
skipBlankInXML noblank = noblank


isAllBlank :: (String -> Bool)
isAllBlank = and . map isOneBlank

isOneBlank :: Char -> Bool
isOneBlank ' ' = True
isOneBlank '\n' = True
isOneBlank '\t' = True
isOneBlank _ = False


takeOneElement :: [FRItem] -> [String] -> (String, String, SaxStream) -> [(String, String)] -> [FRItem]
takeOneElement itemSoFar cateStack (questionPart, answerPart, rest) attrList =
		result rest (FRItem questionPart answerPart attrList cateStack : itemSoFar) cateStack


extractItemPart :: Char -> SaxStream -> (String, SaxStream)
-- suck out the q or a part which is presumed to be at the head of the stream

extractItemPart character ((SaxElementOpen tagOpen []):(SaxCharData text):(SaxElementClose tagClose):rest, whatever) =
	if tagOpen == expectedTag && tagClose == expectedTag then
		(text, skipBlankInXML (rest, whatever))
	else
		error $ "Expected tag <" ++ character : "> not found"
	where
		expectedTag = [character]

-- choice for a null Q or A:
extractItemPart character ((SaxElementOpen tagOpen []):(SaxElementClose tagClose):rest, whatever) =
	if tagOpen == expectedTag && tagClose == expectedTag then
		("", skipBlankInXML (rest, whatever))
	else
		error $ "Expected tag <" ++ character : "> not found"
	where
		expectedTag = [character]
extractItemPart character _ = error $ "Expected " ++ character : " part"


lookUpStrAttr :: String -> [Attribute] -> String
lookUpStrAttr name key = lookUpResult $ lookup name key

extractSingleValue :: AttValue -> String
extractSingleValue (AttValue list) =
	case list of
	[value] ->
		case value of
		Left string ->
			string
		Right _ ->
			error "reference in attr. list"
	_ ->
		error "malformed attribute"

lookUpResult :: (Maybe AttValue) -> String
lookUpResult (Just list) = extractSingleValue list
lookUpResult Nothing = error $ "key not found"
